/*******************************************************************************
 * HMMSegregants - Hidden Markov Model for bulk segregants analysis
 * Copyright 2012 Jorge Duitama
 *
 * This file is part of HMMSegregants.
 *
 *     HMMSegregants is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     HMMSegregants is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with HMMSegregants.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.kuleuven.hmmSegregants;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import JSci.maths.SpecialMath;


public class HMMSegregants {
	private String seqName;
	private List<HMMState> parent1States = new ArrayList<HMMState>();
	private List<HMMState> parent2States = new ArrayList<HMMState>();
	private List<HMMState> backgroundStates = new ArrayList<HMMState>();
	private List<Double> transitions = new ArrayList<Double>();
	
	public HMMSegregants(String seqName) {
		super();
		this.seqName = seqName;
	}
	public static List<HMMSegregants> loadModels (List<HMMVariant> variants, BetaParameter paramsP, BetaParameter paramsB, double r, int numSegregantsN) throws IOException {
		List<HMMSegregants> models = new ArrayList<HMMSegregants>();
		HMMSegregants model = null;
		for(HMMVariant variant: variants) {
			String seqName = variant.getSeqName();
			int position = variant.getPosition();
			int totalCount = variant.getTotalCount();
			int altCount = variant.getAlternativeCount();
			if(model==null || !model.seqName.equals(seqName)) {
				model = new HMMSegregants(seqName);
				models.add(model);
			}
			//Correct for number of segregants
			if(totalCount > numSegregantsN) {
				altCount = (int)Math.round((double)altCount*numSegregantsN/totalCount);
				totalCount = numSegregantsN;
			}
			HMMState sP1 = buildState(variant, paramsP, totalCount, altCount);
			HMMState sP2 = buildState(variant, paramsP, totalCount, totalCount-altCount);
			HMMState sB = buildState(variant, paramsB, totalCount, altCount);
			//if(altCount == totalCount)
			//if(sP.getEmissionLogProb()>sB.getEmissionLogProb())
				//System.out.println("Emmision "+sP.getPosition()+" "+sP.getEmissionLogProb()+" "+sB.getEmissionLogProb());
			//System.out.println("Emission1 : "+ sP1.getEmissionLogProb());
			model.parent1States.add(sP1);
			model.parent2States.add(sP2);
			model.backgroundStates.add(sB);
			if(model.parent1States.size()>1) {
				double distance = position - model.parent1States.get(model.parent1States.size()-2).getPosition();
				double transition = Math.exp(-distance*r);
				//System.out.println("Distance: "+distance+". Transition: "+ transition);
				model.transitions.add(Math.max(0.5, transition));
				//model.transitions.add(0.999);
				//model.transitions.add(0.5);
			}
		}
		
		return models;
	}
	private static HMMState buildState(HMMVariant variant,BetaParameter params, int totalCount, int altCount) {
		double alpha = params.getAlpha();
		double beta = params.getBeta();
		//double beta1 = SpecialMath.beta(alpha + altCount, beta + totalCount - altCount);
		double logBeta1 = SpecialMath.logBeta(alpha + altCount, beta + totalCount - altCount);
		//System.out.println("Beta1: "+beta1);
		//double beta2 = SpecialMath.beta(alpha, beta);
		double logBeta2 = SpecialMath.logBeta(alpha, beta);
		//System.out.println("Beta2: "+beta2);
		//System.out.println("Combinatorial: "+combinatorial);
		double n = totalCount;
		double k = altCount;
		double logCombinatorial = SpecialMath.logGamma(n+1.0)-SpecialMath.logGamma(k+1.0)-SpecialMath.logGamma(n-k+1.0);
		
		double log10combinatorial = logCombinatorial/Math.log(10);
		double log10Beta1 = logBeta1/Math.log(10);
		double log10Beta2 = logBeta2/Math.log(10);
		
		//double combinatorialOld = ExtraMath.binomial((double)totalCount, (double)altCount);
		double emissionLogProb = log10combinatorial + log10Beta1 - log10Beta2;
		//HMMState state = new HMMState(variant, Math.log10(combinatorialOld*beta1/beta2));
		HMMState state = new HMMState(variant, emissionLogProb);

		return state;
	}
	public static List<HMMVariant> loadVariants(String filename) throws IOException {
		List<HMMVariant> variants = new ArrayList<HMMVariant>();
		FileInputStream fis = new FileInputStream(filename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String line=in.readLine();
		String separator = "\t| ";
		while(line!=null) {
			String [] items = line.split(separator);
			String seqName = items[0];
			int position = Integer.parseInt(items[1]);
			int totalCount = Integer.parseInt(items[4]);
			int altCount = Integer.parseInt(items[5]);
			variants.add(new HMMVariant(seqName, position, items[2],items[3], totalCount, altCount));
			line=in.readLine();
		}
		fis.close();
		return variants;
	}
	public void calculateForward() {
		double l1 = Math.log10(1.0/3.0);
		parent1States.get(0).setForwardLogProb(l1);
		parent2States.get(0).setForwardLogProb(l1);
		backgroundStates.get(0).setForwardLogProb(l1);
		for(int i=1;i<parent1States.size();i++) {
			HMMState sP1 = parent1States.get(i-1);
			double lp1 = sP1.getForwardLogProb() + sP1.getEmissionLogProb();
			HMMState sP2 = parent2States.get(i-1);
			double lp2 = sP2.getForwardLogProb() + sP2.getEmissionLogProb();
			HMMState sB = backgroundStates.get(i-1);
			double lb = sB.getForwardLogProb()+sB.getEmissionLogProb();
			double transitionEq = transitions.get(i-1);
			double ltEq = Math.log10(transitionEq);
			double ltD = Math.log10((1-transitionEq)/2);
			double sTemp = logSum(lp1+ltEq,lp2+ltD);
			parent1States.get(i).setForwardLogProb(logSum(sTemp, lb+ltD));
			sTemp = logSum(lp1+ltD,lp2+ltEq);
			parent2States.get(i).setForwardLogProb(logSum(sTemp, lb+ltD));
			sTemp = logSum(lp1+ltD,lp2+ltD);
			backgroundStates.get(i).setForwardLogProb(logSum(sTemp,lb+ltEq));
			//System.out.println("ForwardP1 : "+ parent1States.get(i).getForwardLogProb());
			//System.out.println("ForwardP2 : "+ parent2States.get(i).getForwardLogProb());
			//System.out.println("ForwardB : "+ backgroundStates.get(i).getForwardLogProb());
		}
	}
	public double logSum (double log1,double log2) {
		return log1 + Math.log10(1+Math.pow(10.0, log2-log1));
	}
	public void calculateBackward() {
		int last = parent1States.size()-1;
		parent1States.get(last).setBackwardLogProb(0.0);
		parent2States.get(last).setBackwardLogProb(0.0);
		backgroundStates.get(last).setBackwardLogProb(0.0);
		for(int i=last-1;i>=0;i--) {
			HMMState sP1 = parent1States.get(i+1);
			double lp1 = sP1.getBackwardLogProb()+sP1.getEmissionLogProb();
			HMMState sP2 = parent2States.get(i+1);
			double lp2 = sP2.getBackwardLogProb()+sP2.getEmissionLogProb();
			HMMState sB = backgroundStates.get(i+1);
			double lb = sB.getBackwardLogProb()+sB.getEmissionLogProb();
			double transitionEq = transitions.get(i);
			double ltEq = Math.log10(transitionEq);
			double ltD = Math.log10((1-transitionEq)/2);
			double sTemp = logSum(lp1+ltEq,lp2+ltD);
			parent1States.get(i).setBackwardLogProb(logSum(sTemp, lb+ltD));
			sTemp = logSum(lp1+ltD,lp2+ltEq);
			parent2States.get(i).setBackwardLogProb(logSum(sTemp, lb+ltD));
			sTemp = logSum(lp1+ltD,lp2+ltD);
			backgroundStates.get(i).setBackwardLogProb(logSum(sTemp,lb+ltEq));
			//System.out.println("BackwardP1 : "+ parent1States.get(i).getBackwardLogProb());
			//System.out.println("BackwardP2 : "+ parent2States.get(i).getBackwardLogProb());
			//System.out.println("BackwardB : "+ backgroundStates.get(i).getBackwardLogProb());
		}
	}
	public List<HMMVariant> getBackgroundVariants() {
		List<HMMVariant> answer = new ArrayList<HMMVariant>();
		for(int i=0;i<parent1States.size();i++) {
			HMMState sP1 = parent1States.get(i);
			HMMState sP2 = parent2States.get(i);
			HMMState sB = backgroundStates.get(i);
			double lp1 = sP1.getForwardLogProb()+sP1.getBackwardLogProb()+sP1.getEmissionLogProb();
			double lp2 = sP2.getForwardLogProb()+sP2.getBackwardLogProb()+sP2.getEmissionLogProb();
			double lb = sB.getForwardLogProb()+sB.getBackwardLogProb()+sB.getEmissionLogProb();
			if(lb>lp1 && lb>lp2) {
				answer.add(sB.getVariant());
			}
		}
		return answer;
	}
	public void saveResults (BetaParameter paramsP) {
		for(int i=0;i<parent1States.size();i++) {
			HMMState sP1 = parent1States.get(i);
			HMMState sP2 = parent2States.get(i);
			HMMState sB = backgroundStates.get(i);
			HMMVariant variant = sP1.getVariant();
			double lp1 = sP1.getForwardLogProb()+sP1.getBackwardLogProb()+sP1.getEmissionLogProb();
			double lp2 = sP2.getForwardLogProb()+sP2.getBackwardLogProb()+sP2.getEmissionLogProb();
			double lb = sB.getForwardLogProb()+sB.getBackwardLogProb()+sB.getEmissionLogProb();
			double lMax = Math.max(lp1, lp2);
			lMax = Math.max(lMax, lb);
			double lp1M = lp1 - lMax;
			double lp2M = lp2 - lMax;
			double lbM = lb - lMax;
			double np1 = Math.pow(10.0, lp1M);
			double np2 = Math.pow(10.0, lp2M);
			double nb = Math.pow(10.0, lbM);
			double pp1 = np1/(np1+np2+nb);
			double pp2 = np2/(np1+np2+nb);
			double p = pp1;
			if(pp2>pp1) {
				p = -pp2;
			}
			double absP = Math.abs(p);
			if(variant.getAlpha()==0 || (absP >=0.95 && paramsP.getAlpha() > variant.getAlpha()) ) {
				variant.setLogPosteriorP1(lp1);
				variant.setLogPosteriorP2(lp2);
				variant.setLogPosteriorB(lb);
				variant.setPosteriorProbability(p);
				variant.setAlpha(paramsP.getAlpha());
			}		
		}
	}
	public static void printResults (List<HMMVariant> variants, PrintStream out, boolean printAlpha) {
		for(HMMVariant variant:variants) {
			
			out.print(variant.getPosition());
			//out.print(variant.getSeqName()+"\t"+variant.getPosition()+"\t"+variant.getReferenceAllele()+"\t"+variant.getAlternativeAllele());
			//out.print("\t"+variant.getTotalCount()+"\t"+variant.getAlternativeCount());
			//out.print("\t"+variant.getLogPosteriorP1()+"\t"+variant.getLogPosteriorP2()+"\t"+variant.getLogPosteriorB());
			out.print("\t"+variant.getPosteriorProbability());
			//if(printAlpha) out.print("\t"+variant.getAlpha());
			out.println();
		}
		/*
		System.out.println("36376 : ");
		HMMVariant variant = variants.get(36376);
		out.print(variant.getSeqName()+"\t"+variant.getPosition()+"\t"+variant.getReferenceAllele()+"\t"+variant.getAlternativeAllele());
		out.print("\t"+variant.getTotalCount()+"\t"+variant.getAlternativeCount());
		out.print("\t"+variant.getLogPosteriorP1()+"\t"+variant.getLogPosteriorP2()+"\t"+variant.getLogPosteriorB());
		out.print("\t"+variant.getPosteriorProbability());
		if(printAlpha) out.print("\t"+variant.getAlpha());
		out.println();
		*/
	}
	public static void main(String[] args) throws Exception {
		int i=0;
		double alphaP = -1;
		double betaP = 1;
		double r = 3.5 * Math.pow(10, -6);
		int numSegregantsN = 10000;
		while(i<args.length && args[i].charAt(0)=='-') {
			if("-a".equalsIgnoreCase(args[i])) {
				i++;
				alphaP = Double.parseDouble(args[i]);
			} else if("-b".equalsIgnoreCase(args[i])) {
				i++;
				betaP = Double.parseDouble(args[i]);
			} else if("-r".equalsIgnoreCase(args[i])) {
				i++;
				r = Double.parseDouble(args[i]);
			} else if("-s".equalsIgnoreCase(args[i])) {
				i++;
				numSegregantsN = Integer.parseInt(args[i]);
			}
			i++;
		}
		
		List<HMMVariant> variants = loadVariants(args[i++]);
		BetaParametersEstimator estimator = new BetaParametersEstimator();
		BetaParameter paramsB = estimator.estimateParameters(variants);
		System.err.println("Alpha background: "+paramsB.getAlpha()+" Beta background: "+paramsB.getBeta());
		double alphaPMin = 5;
		double alphaPMax = 50;
		if(alphaP>0) {
			alphaPMin = alphaPMax = alphaP;
		}
		for(double a=alphaPMin;a<=alphaPMax;a++) {
			BetaParameter paramsP = new BetaParameter(a, betaP);
			System.err.println("Alpha phenotype: "+paramsP.getAlpha()+" Beta phenotype: "+paramsP.getBeta());
			int nIters = 5;
			for(int iter=0;iter < nIters;iter++) {
				List<HMMSegregants> models = loadModels(variants,paramsP,paramsB,r, numSegregantsN);
				List<HMMVariant> variantsB = new ArrayList<HMMVariant>();
				for(HMMSegregants hmm:models) {
					hmm.calculateForward();
					hmm.calculateBackward();
					if (iter == nIters-1) {
						hmm.saveResults(paramsP);
					} else if(a == alphaPMin) {
						variantsB.addAll(hmm.getBackgroundVariants());
					}
				}
				if(variantsB.size()>30) {
					paramsB = estimator.estimateParameters(variantsB);
					System.err.println("Alpha background: "+paramsB.getAlpha()+" Beta background: "+paramsB.getBeta());
				}
			}
		}
		printResults(variants, System.out, alphaP<=0);	
	}
}
