package edu.kuleuven.hmmSegregants;

import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.TreeMap;
import java.util.TreeSet;

public class SegregantsSimulator {
	
	protected int nSites=750000;
	protected int nMarkerSites=2500;
	protected int nCausativeSites=1;
	protected int nTotalSegregants=200;
	protected int nSelectedSegregants=50;
	protected int coverage=100;
	//protected double recombinationRate = 3.5 * Math.pow(10, -6);
	protected double cmPerKbp = 0.37;
	private double avgQTLEffect=0.9;
	private double stdevQTLEffect=0.05;
	protected double sequencingErrorRate=0.01;
	protected int model = 0;
	protected String outPrefix;
	
	Map<Integer, Double> causalIndexes;
	
	protected static final Random random = new Random(System.currentTimeMillis());
	
	public static void main(String[] args) throws Exception {
		SegregantsSimulator simulator = new SegregantsSimulator();
		int i=0;
		while(i<args.length && args[i].charAt(0)=='-') {
			if("-S".equals(args[i])) {
				i++;
				simulator.nSites = Integer.parseInt(args[i]);
			} else if("-s".equals(args[i])) {
				i++;
				simulator.nCausativeSites = Integer.parseInt(args[i]);
			} else if("-N".equals(args[i])) {
				i++;
				simulator.nTotalSegregants = Integer.parseInt(args[i]);
			} else if("-MS".equals(args[i])) {
				i++;
				simulator.nMarkerSites = Integer.parseInt(args[i]);
			} else if("-n".equals(args[i])) {
				i++;
				simulator.nSelectedSegregants = Integer.parseInt(args[i]);
			} else if("-c".equals(args[i])) {
				i++;
				simulator.coverage = Integer.parseInt(args[i]);
			} else if("-r".equals(args[i])) {
				i++;
				//simulator.recombinationRate = Double.parseDouble(args[i]);
				simulator.cmPerKbp = Double.parseDouble(args[i]);
			} else if("-q".equals(args[i])) {
				i++;
				simulator.avgQTLEffect = Double.parseDouble(args[i]);
			} else if("-d".equals(args[i])) {
				i++;
				simulator.stdevQTLEffect = Double.parseDouble(args[i]);
			} else if("-e".equals(args[i])) {
				i++;
				simulator.sequencingErrorRate = Double.parseDouble(args[i]);
			} else if("-m".equals(args[i])) {
				i++;
				simulator.model = Integer.parseInt(args[i]);
			}
			i++;
		}
		if (args.length > 0)
			simulator.outPrefix = args[i];
		else 
			simulator.outPrefix = "test";
		simulator.printParameters(System.out);
		simulator.simulate();
	}
	
	protected void printParameters(PrintStream out) {
		out.println("Total number of sites: "+nSites);
		out.println("Snp Density: "+nMarkerSites);
		out.println("Number of causative sites: "+nCausativeSites);
		out.println("Total number of segregants: "+nTotalSegregants);
		out.println("Number of segregants to select: "+nSelectedSegregants);
		out.println("Average coverage: "+coverage);
		//out.println("Recombination rate between sites: "+recombinationRate);
		out.println("Recombination rate (cM/kbp): "+cmPerKbp);
		out.println("Average QTL effect: "+avgQTLEffect);
		out.println("Standard deviation QTL effect: "+stdevQTLEffect);
		out.println("Sequencing error rate: "+sequencingErrorRate);
		if(model==0) out.println("Model: Quantitative selection");
		if(model==1) out.println("Model: Survival additive selection");
		if(model==2) out.println("Model: Survival all sites selection");
		if(model==3) out.println("Model: Survival at least one site selection");
		
	}

	public void simulate() throws IOException {
		System.out.println("Making segregants");
		TreeSet<Integer> snps = makeSNPs();
		if (snps.size() != nMarkerSites)
			throw new RuntimeException();
		
		boolean [][] segregants = makeSegregants(snps);
		System.out.println("Simulating counts unselected");
		int [][] counts = makeCounts (segregants);
		System.out.println("Printing unselected counts");
		printCounts(counts, snps, outPrefix+"_unselected.counts");
		System.out.println("Choosing causal sites");
		causalIndexes = chooseCausalIndexes();
		printCausalSites(causalIndexes, snps ,outPrefix+".sites");
		System.out.println("Selecting segregants");
		boolean [][] selectedSeg = select(segregants, causalIndexes,true);
		System.out.println("Simulating counts");
		counts = makeCounts (selectedSeg);
		System.out.println("Printing counts");
		printCounts(counts, snps, outPrefix+"_high.counts");
		if(model == 0) {
			System.out.println("Selecting low tail segregants");
			selectedSeg = select(segregants, causalIndexes,false);
			System.out.println("Simulating low tail counts");
			counts = makeCounts (selectedSeg);
			System.out.println("Printing low tail counts");
			printCounts(counts, snps, outPrefix+"_low.counts");
		}
	}

	protected TreeSet<Integer> makeSNPs() {
		TreeSet<Integer> snps = new TreeSet<Integer>();
		while(snps.size() < nMarkerSites) {
			snps.add(random.nextInt(nSites));
		}
		return snps;
	}

	protected boolean [][] makeSegregants(TreeSet<Integer> snps) {
		
		boolean [][] segregants = new boolean[nTotalSegregants][];
		for(int i=0; i<nTotalSegregants;i++) {
			segregants[i] = makeSegregant(snps);
		}
		
		/*
		for(int i=0; i<nTotalSegregants;i++) {
			boolean nextCall = false;
			
			Iterator<Integer> snps_it = snps.iterator();
			int next_snp = snps_it.next();
			
			int count = 0;
			for(int j = 0; j < nSites && snps_it.hasNext(); j++) {

				double changeProb = j > 0? recombinationRate : 0.5;
				if(random.nextDouble() < changeProb)
					nextCall = !nextCall;
				
				if (j == next_snp) {
					segregants[i][count] = nextCall;
					next_snp = snps_it.next();
					count++;
				}
			}
		}
		*/
		return segregants;
	}
	
	protected boolean [] makeSegregant(TreeSet<Integer> snps) {
		boolean [] segregant = new boolean[nMarkerSites];
		Iterator<Integer> snps_it = snps.iterator();

		boolean nextCall = false;
		
		int last_snp = 0;
		
		for(int j = 0; j < nMarkerSites; j++) {

			int snp = snps_it.next();
			
			if (j == 0) {
				nextCall = random.nextDouble() < 0.5;
				
			} else {
				
				double distance_bp = snp - last_snp;
				/*
				 * Haldane:
				 * 
				 * d = 1/2 * ln(1 - 2r)
				 * r = 1/2 * (1 - e^(-2d)) *
				 * 
				 */
				
				double distance_cM = cmPerKbp * distance_bp / 1000;
				double distance_M = distance_cM / 100;
				double r = 0.5 * (1 - Math.exp(-2*distance_M));
				
				if(random.nextDouble() < r) 
					nextCall = !nextCall;
				
			}
			segregant[j] = nextCall;
			
			last_snp = snp;
		}
		return segregant;
	}

	protected Map<Integer, Double> chooseCausalIndexes() {
		Map<Integer, Double> causalIndexes = new TreeMap<Integer, Double>();
		for(int i=0;i<nCausativeSites;i++) {
			int nextSite = random.nextInt(nMarkerSites);
			double nextEffect = avgQTLEffect;
			if(nCausativeSites>1) nextEffect = random.nextGaussian()*stdevQTLEffect+avgQTLEffect;
			if(nextEffect>0.99) nextEffect=0.99;
			else if (nextEffect<0.51) nextEffect = 0.51;
			causalIndexes.put(nextSite, nextEffect);
		}
		return causalIndexes;
	}

	protected boolean[][] select(boolean[][] segregants,Map<Integer, Double> causalIndexes, boolean highTail) {
		//Calculate score per segregant
		List<SegregantScore> scores = new ArrayList<SegregantScore>();
		for(int i=0;i<segregants.length;i++) {
			double scoreSum = 0;
			double scoreMax = 0;
			double scoreMin = 1;
			for(Map.Entry<Integer, Double> entry:causalIndexes.entrySet()) {
				boolean value = segregants[i][entry.getKey()];
				double contribution = value? entry.getValue() : 1-entry.getValue();
				scoreSum+=contribution;
				if(scoreMax<contribution) scoreMax = contribution;
				if(scoreMin>contribution) scoreMin = contribution;
			}
			if(model==3) {
				scores.add(new SegregantScore(i, scoreMax));
			} else if (model == 2) {
				scores.add(new SegregantScore(i, scoreMin));
			} else if (model == 1) {
				scores.add(new SegregantScore(i, scoreSum/causalIndexes.size()));
			} else {
				scores.add(new SegregantScore(i, scoreSum + random.nextGaussian()*stdevQTLEffect));
			}
			
		}
		
		List<SegregantScore> filteredScores = new ArrayList<SegregantScore>();
		
		//Choose the segregants that will survive
		List<SegregantScore> surviveScores = new ArrayList<SegregantScore>();
		for(SegregantScore sc:scores) {
			double survivalProb = sc.getScore();
			//System.out.println("Segregant: "+sc.getIndex()+". Survival prob: "+survivalProb);
			if(random.nextDouble()<survivalProb) surviveScores.add(sc);
		}
		
		if(model >0 && model <=3) {
			//Since the segregants are not sorted, just choose the first n for the final group
			for(int i=0;i<nSelectedSegregants && i<surviveScores.size();i++) filteredScores.add(surviveScores.get(i));
		} else {
			//Choose the best scores in the distribution
			Collections.sort(surviveScores,new SegregantsScoreComparator());
			if(highTail) Collections.reverse(surviveScores);
			for(int i=0;filteredScores.size() < nSelectedSegregants;i++) {
				filteredScores.add(surviveScores.get(i));
			}
		}
		
		boolean[][] selectedSegregants = new boolean[filteredScores.size()][segregants[0].length];
		for(int i=0;i<selectedSegregants.length;i++) {
			SegregantScore sc = filteredScores.get(i);
			int k = sc.getIndex();
			for(int j=0;j<selectedSegregants[0].length;j++) {
				selectedSegregants[i][j] = segregants[k][j];
			}
		}
		return selectedSegregants;
	}

	protected int[][] makeCounts(boolean[][] selectedSeg) {
		int [][] counts = new int[2][nMarkerSites];

		/*
		PoissonDistribution c = new PoissonDistribution(coverage);
		for (int snp = 0; snp < nMarkerSites; snp ++) {
			int specific_coverage = (int) c.inverse(random.nextDouble());
			for (int r = 0; r < specific_coverage; r ++) {
				int i = random.nextInt(selectedSeg.length);
				boolean val = selectedSeg[i][snp];
				if(random.nextDouble() < sequencingErrorRate)
					val = !val;
				
				counts[val? 1 : 0][snp]++;
				
			}
		}
		*/
		
		int nReads = nMarkerSites*coverage;
		for(int k=0; k < nReads;k++) {
			int i = random.nextInt(selectedSeg.length);
			int j = random.nextInt(nMarkerSites);
			boolean val = selectedSeg[i][j];
			if(random.nextDouble() < sequencingErrorRate)
				val = !val;
			counts[val ? 0:1][j]++;
		}
		return counts;
	}

	protected void printCausalSites(Map<Integer, Double> causalIndexes, TreeSet<Integer> snps, String outFile)  throws IOException {
		PrintStream outSites = new PrintStream(outFile);
		for(int i:causalIndexes.keySet()) {
			int realMarkerSite = getMarkerSite(snps, i);
			outSites.println(""+realMarkerSite+"\t"+causalIndexes.get(i));
		}
		outSites.flush();
		outSites.close();
		
	}
	
	protected int getMarkerSite(TreeSet<Integer> snps, int nextInt) {
		
		int count = 0;
		
		for (Iterator<Integer> snps_it = snps.iterator(); snps_it.hasNext(); count++) {
			int marker = snps_it.next();
			if (count == nextInt)
				return marker;
		}
		// should find the marker
		throw new RuntimeException();
	}

	protected void printCounts(int[][] counts, TreeSet<Integer> snps, String outFile) throws IOException  {
		PrintStream outCounts = new PrintStream(outFile);
		
		Iterator<Integer> snps_it = snps.iterator();
				
		for(int j=0; j < nMarkerSites && snps_it.hasNext(); j++) {
			int total = counts[0][j]+counts[1][j];
			int pos = snps_it.next();
			outCounts.println("ChrSim\t"+pos+"\t0\t1\t"+total+"\t"+counts[1][j]);
		}
		outCounts.flush();
		outCounts.close();
	}
}
class SegregantScore {
	private int index;
	private double score;
	public SegregantScore(int index, double score) {
		super();
		this.index = index;
		this.score = score;
	}
	public int getIndex() {
		return index;
	}
	public double getScore() {
		return score;
	}
	@Override
	public String toString() {
		return index+"["+score+"]";
	}
	
}
class SegregantsScoreComparator implements Comparator<SegregantScore> {

	@Override
	public int compare(SegregantScore s1, SegregantScore s2) {
		if(s1.getScore()<s2.getScore())return -1;
		else if (s1.getScore()>s2.getScore()) return 1;
		else return 0;
	}
}
